import os
import shutil

from ..util import get_files
from .base import Backend


def ensure_directory(fp):
    fp = os.path.abspath(fp)
    if not os.path.isdir(fp):
        os.makedirs(fp)
    return fp


class FilesystemBackend(Backend):
    def get_base_path(self):
        return ensure_directory(self.data_root)

    def exists(self, path):
        p = self.get_path(path)
        return any((os.path.isfile(p), os.path.isdir(p)))

    def save(self, path, content):
        path = self.get_path(path)
        ensure_directory(os.path.split(path)[0])
        with open(path, "w") as f:
            content = str(content)
            f.write(content)
        return path

    def _load(self, path):
        path = self.get_path(path)
        with open(path) as f:
            content = f.read().strip()
        return content

    def get_children(self, path=".", condition=lambda x: True):
        path = self.get_path(path)
        return get_files(path, condition)

    def delete(self, path=""):
        path = self.get_path(path)
        if os.path.isdir(path) and not os.path.islink(path):
            shutil.rmtree(path)
        elif os.path.exists(path):
            os.remove(path)



import os
from datetime import date, datetime
from hashlib import sha1
from pathlib import Path

# from banal import as_bool, clean_dict

BUF_SIZE = 1024 * 1024 * 16
HASH_LENGTH = 40  # sha1


def get_files(directory, condition=lambda x: True):
    """
    yield tuples of (filename, path) for files in given `directory`
    that match `condition` (default: all) incl. subdirectories
    """
    return (
        (os.path.splitext(f)[0], os.path.join(d, f))
        for d, _, fnames in os.walk(directory)
        for f in fnames
        if condition(os.path.join(d, f))
    )


def cast(value, with_date=False):
    if not isinstance(value, (str, float, int)):
        return value
    if isinstance(value, str):
        value = value.strip()
        if not value:  # ''
            return None
    try:
        if float(value) == int(float(value)):
            return int(value)
        return float(value)
    except (TypeError, ValueError):
        if with_date:
            try:
                return datetime.fromisoformat(value)
            except ValueError:
                pass
        # value = as_bool(value, None)
        return value


def flatten_dict(d):
    def items():
        for key, value in d.items():
            key = key.replace("-", "_")
            if isinstance(value, dict):
                for subkey, subvalue in flatten_dict(value).items():
                    yield key + ":" + subkey, subvalue
            else:
                yield key, value

    return dict(items())


def casted_dict(d, ignore_keys=[]):
    return {
        k: cast(v, with_date=True) if k not in ignore_keys else v for k, v in d.items()
    }


def robust_dict(d):
    # no typing for better comparison performance
    return {k: str(v) if v else None for k, v in flatten_dict(d).items()}


def ensure_path(file_path):
    if file_path is None or isinstance(file_path, Path):
        return file_path
    return Path(file_path).resolve()


def checksum(file_name):
    """Generate a hash for a given file name."""
    file_name = ensure_path(file_name)
    if file_name is not None and file_name.is_file():
        digest = sha1()
        with open(file_name, "rb") as fh:
            while True:
                block = fh.read(BUF_SIZE)
                if not block:
                    break
                digest.update(block)
        return str(digest.hexdigest())


def dict_diff(dict1, dict2):
    """
    return key/value pairs from dict1 that are different from dict2
    """
    return set(flatten_dict(dict1).items()) - set(flatten_dict(dict2).items())


def dict_is_subset(dict1, dict2, ignore=set()):
    """
    check if dict1 is contained in dict2 (including values)
    """
    return len(dict_diff(dict1, dict2) - ignore) == 0


def datetime_to_json(value):
    if isinstance(value, date):
        return value.isoformat()



import json
import os

from ..util import cast, datetime_to_json


class Backend:
    """
    base class for metadir backends.
    currently only local filesystem backend implemented.
    """

    def __init__(self, data_root):
        self.data_root = data_root
        self.base_path = self.get_base_path()

    def __str__(self):
        return self.get_base_path()

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self}>"

    def get_base_path(self):
        """return a base path to a local file dir or a cloud bucket"""
        raise NotImplementedError

    def get_path(self, path):
        """return absolute filesystem path or cloud bucket for `path"""
        return os.path.join(self.base_path, path)

    def exists(self, path):
        """check if given path exists and return boolean"""
        raise NotImplementedError

    def save(self, path, content):
        """
        store `content` in path and return absolute path to stored file or
        cloud blob location
        """
        raise NotImplementedError

    def load(self, path):
        """
        return content as string for given path, use the same not found
        exception for all storages:
        """
        if not self.exists(path):
            raise FileNotFoundError(f"Path `{path}` not found in storage `{self}`")
        return self._load(path)

    def load_json(self, path):
        return json.loads(self.load(path))

    def dump_json(self, path, content):
        content = json.dumps(content, default=datetime_to_json)
        self.save(path, content)

    def _load(self, path):
        """actual implementation for specific storage"""
        raise NotImplementedError

    def set_value(self, path, value):
        """simply store values to a path location"""
        self.save(path, value)
        return value

    def get_value(self, path, transform=lambda x: cast(x, with_date=True)):
        """simply get values from a path location"""
        if not self.exists(path):
            return
        content = self.load(path)
        return transform(content)

    def get_children(self, path=".", condition=lambda x: True):
        """list all children under given path that match condition"""
        raise NotImplementedError

    def delete(self, path=""):
        """delete everything from path"""
        raise NotImplementedError



import csv
import os
import sys

import dataset
from sqlalchemy.sql import func

from . import settings
from .backend.appendonly import AppendOnlyBackend
from .backend.filesystem import FilesystemBackend
from .backend.store import Store
from .config import Config
from .db import generate_metadata, update_state_db
from .file import FilesWrapper


class Metadir:
    def __init__(self, base_path=None, files_root=None):
        self._base_path = base_path or settings.MMMETA
        self._files_root = files_root or base_path or settings.MMMETA_FILES_ROOT
        self._backend = FilesystemBackend(os.path.join(self._base_path, "_mmmeta"))
        self.config = Config(self)
        self._metadata = AppendOnlyBackend(
            self._backend.get_path("db"), self.config.unique
        )
        self._db_path = f'sqlite:///{self._backend.get_path("state.db")}'
        self.store = Store(FilesystemBackend(self._backend.get_path("_store")))

    def __repr__(self):
        return f"<Metadir: `{self._backend.__class__.__name__}` {self._backend}>"

    def __len__(self):
        return len(self.files)

    @property
    def files(self):
        return FilesWrapper(self._db["files"], self)

    @property  # Shorthand
    def _db(self):
        return dataset.connect(self._db_path)

    def generate(
        self,
        path=None,
        replace=False,
        ensure_metadata=False,
        ensure_files=False,
        no_meta=False,
    ):
        """
        generate or update metadata
        """
        backend = FilesystemBackend(path or self._files_root)
        return generate_metadata(
            backend, self, replace, ensure_metadata, ensure_files, no_meta
        )

    def squash(self):
        self._metadata.squash()

    def update(self, replace=False, cleanup=False):
        """
        update local state with meta db
        """
        return update_state_db(self, replace, cleanup)

    def inspect(self):
        """
        return some insights
        """
        return {
            "files": len(self),
            "path": str(self._backend),
        }

    def touch(self, key):
        """
        store a timestamp with given key
        """
        return self.store.touch(key)

    def dump(self, out=sys.stdout):
        """
        dump csv
        """
        file = self.files.find_one()
        columns = set(file._data.keys()) | set(vars(file.remote).keys())
        writer = csv.DictWriter(out, fieldnames=columns)
        writer.writeheader()
        for file in self.files:
            writer.writerow(file.serialize())

    @property
    def state_last_updated(self):
        table = self.files._table.table
        query = func.max(table.c["__state_last_updated"])
        for res in self._db.query(query):
            return res.get("max_1")

    @property
    def meta_last_updated(self):
        table = self.files._table.table
        query = func.max(table.c["__meta_last_updated"])
        for res in self._db.query(query):
            return res.get("max_1")

    @property
    def last_touched(self):
        return max(
            self.store["store_last_updated"],
            self.state_last_updated,
            self.meta_last_updated,
        )



import yaml
from banal import ensure_dict, ensure_list


class Config:
    """
    metadata:
      file_name: _file_name
      required:
      - published_at
      - foreign_id
      include:
      - reference
      - modified_at
      - title
      - originators
      - publisher:name
      unique: content_hash
    remote:
      url: https://my_bucket.s3.eu-central-1.amazonaws.com/foo/bar/{_file_name}
      uri: s3://my_bucket/foo/bar/{_file_name}
    """

    def __init__(self, m):
        self._m = m
        self._config = {}
        config_path = m._backend.get_path("config.yml")
        if m._backend.exists(config_path):
            with open(config_path) as f:
                config = yaml.load(f)
            self._config = ensure_dict(config)
        self._metadata = ensure_dict(self["metadata"])
        self._remote = ensure_dict(self["remote"])

    def __getitem__(self, item):
        return self._config.get(item)

    def __contains__(self, item):
        return item in self._config

    @property
    def unique(self):
        return self._metadata.get("unique", "content_hash")

    @property
    def file_name(self):
        return self._metadata.get("file_name", self.unique)

    @property
    def required_keys(self):
        def _get_required_keys():
            yield self.unique
            yield self.file_name
            meta = ensure_dict(self["metadata"])
            if "required" in meta:
                for key in ensure_list(meta["required"]):
                    yield key

        return set(_get_required_keys())

    @property
    def keys(self):
        def _get_keys():
            yield from self.required_keys
            for key in ensure_list(self._metadata.get("include")):
                yield key

        return set(_get_keys())

    def get_remote(self, data):
        """
        compute a remote url or uris with simple string replacement from
        `remote` config
        """
        for key, value in self._remote.items():
            yield key, value.format(**data)



import os


def get_env(key, default=None):
    return os.environ.get(key, default)


MMMETA = os.path.abspath(get_env("MMMETA", os.getcwd()))
MMMETA_FILES_ROOT = os.path.abspath(get_env("MMMETA_FILES_ROOT", MMMETA))

LOGGING = get_env("LOGGING")
LOG_FORMAT = get_env("LOG_FORMAT", "TEXT")



import csv
from datetime import datetime

import dataset

from ..util import robust_dict
from .filesystem import FilesystemBackend, ensure_directory


class AppendOnlyBackend(FilesystemBackend):
    def __init__(self, base_path, unique):
        super().__init__(base_path)
        self.unique = unique

    def _get_table(self, tx, name="tmp"):
        return tx.get_table(name, primary_id=self.unique, primary_type=tx.types.text)

    def write(self, table, suffix="append"):
        fp = self.get_path(datetime.now().isoformat() + f".{suffix}")
        ensure_directory(self.base_path)  # FIXME
        if hasattr(table, "all"):  # FIXME
            table = table.all()
        try:
            # maybe we have 0 rows:
            data = next(table)
            with open(fp, "w") as f:
                writer = csv.DictWriter(f, fieldnames=data.keys())
                writer.writeheader()
                writer.writerow(data)
                for data in table:
                    writer.writerow(data)
        except StopIteration:
            pass

    def load_step(self, path):
        path = self.get_path(path)
        with open(path) as f:
            reader = csv.DictReader(f)
            yield from reader

    def load(self, table):
        """
        files are named by timestamp, so we can order the history
        walk up until the most recent squashed, then walk back down
        from there to generate data
        """

        def _get_steps():
            steps = reversed(sorted(c[1] for c in self.get_children()))
            for step in steps:
                if step.endswith("append"):
                    yield step
                if step.endswith("squashed"):
                    yield step
                    break

        for step in reversed(list(_get_steps())):
            for row in self.load_step(step):
                row = robust_dict(row)
                keys = row.get("__mmmeta_keys")
                if keys:
                    keys = keys.split(",") + [self.unique]
                    row = {k: v for k, v in row.items() if k in keys}
                table.upsert(row, [self.unique])

    def squash(self):
        with dataset.connect("sqlite:///:memory:") as tx:
            table = self._get_table(tx)
            self.load(table)
            self.write(table, "squashed")



import os
from datetime import datetime
from types import SimpleNamespace

from banal import clean_dict

from .exceptions import ValidationError


class File:
    def __init__(self, metadir, data):
        self._metadir = metadir
        self._data = data
        self._unique = metadir.config.unique

    def __setitem__(self, attr, value):
        self.update(**{attr: value})

    def __getitem__(self, attr, default=None):
        return self._data.get(attr, default)

    def __contains__(self, key):
        return key in self._data

    def update(self, **data):
        """
        bulk attribute update (like dict.update)
        update internal data object and write to state db
        """
        self._data.update(**data)
        self._data["__state_last_updated"] = datetime.now()

    def save(self):
        with self._metadir._db as db:
            db["files"].update(self._data, [self._unique])

    def serialize(self):
        return {**self._data, **vars(self.remote)}

    @property
    def uid(self):
        return self._data[self._unique]

    @property
    def name(self):
        return self._data.get(self._metadir.config.file_name, self.uid)

    @property
    def remote(self):
        return SimpleNamespace(**dict(self._metadir.config.get_remote(self._data)))


class FilesWrapper:
    """
    yield actual `File` objects from dataset table,
    pass other operations through `dataset.table.Table`

    https://dataset.readthedocs.io/en/latest/quickstart.html#reading-data-from-tables
    """

    def __init__(self, table, metadir):
        self._table = table
        self._metadir = metadir
        self.config = metadir.config

    def __iter__(self):
        for data in self._table:
            yield File(self._metadir, data)

    def __len__(self):
        return len(self._table)

    def __contains__(self, file):
        return bool(self.find_one(**{self.config.unique: file[self.config.unique]}))

    def find(self, *args, **kwargs):
        for data in self._table.find(*args, **kwargs):
            yield File(self._metadir, data)

    def find_one(self, *args, **kwargs):
        data = self._table.find_one(*args, **kwargs)
        if data:
            return File(self._metadir, data)

    def __getattr__(self, attr):
        """
        pass through dataset table funcionality
        """
        return getattr(self._table, attr)

    def validate(self, data):
        """
        check if data dict has all required keys from config
        """
        if hasattr(data, "_data"):
            data = data._data
        data = clean_dict(data)
        remaining = self.config.required_keys - set(data.keys())
        if remaining:
            raise ValidationError(f"Missing keys: {remaining}")
        return True

    def ensure(self, data):
        """
        ensure that an actual file (local storage only) really exists
        """
        if hasattr(data, "_data"):
            data = data._data
        data = clean_dict(data)
        fp = os.path.join(
            self._metadir._files_root, data[self._metadir.config.file_name]
        )
        return os.path.exists(fp)



from datetime import datetime

from ..exceptions import StoreError


class Store:
    """
    simple, file-system based key-value store
    """

    def __init__(self, backend):
        self._backend = backend

    def __str__(self):
        return str(self._backend)

    def __repr__(self):
        return repr(self._backend)

    def __getitem__(self, attr):
        if "/" in attr:
            raise StoreError(f"illegal key: {attr}")
        return self._backend.get_value(attr)

    def __setitem__(self, attr, value=""):
        if "/" in attr:
            raise StoreError(f"illegal key: {attr}")
        self._backend.set_value(attr, value)
        self.touch()

    def __iter__(self):
        for key, _ in self._backend.get_children():
            yield key, self[key]

    def touch(self, key="store_last_updated"):
        self._backend.set_value(key, datetime.now().isoformat())
        if key != "store_last_updated":
            # of course we like to touch always everything
            self._backend.set_value("store_last_updated", datetime.now().isoformat())

    def serialize(self):
        return {k: v for k, v in self}

    def to_string(self):
        tmpl = "{k}: {v}"
        return "\n".join(tmpl.format(k=k, v=v) for k, v in self)



from setuptools import setup, find_packages


def readme():
    with open("README.rst") as f:
        return f.read().strip()


setup(
    name="mmmeta",
    version="0.4.3",
    description="A simple toolkit for managing local state against remote metadata.",
    long_description=readme(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
    ],
    url="https://github.com/simonwoerpel/mmmeta",
    author="Simon Wörpel",
    author_email="simon.woerpel@medienrevolte.de",
    license="MIT",
    packages=find_packages(exclude=["mmmeta.tests"]),
    package_dir={"mmmeta": "mmmeta"},
    entry_points={"console_scripts": ["mmmeta=mmmeta.cli:cli"]},
    install_requires=[
        "click",
        "dataset",
        "banal",
        "pyyaml",
        "python-dateutil",
        "structlog",
    ],
    zip_safe=False,
)



import logging

import click

from mmmeta import mmmeta, settings
from mmmeta.logging import configure_logging

log = logging.getLogger(__name__)


@click.group()
@click.option(
    "--metadir",
    default=settings.MMMETA,
    help="Base path for reading meta info and storing state",
    show_default=True,
)
@click.option(
    "--files-root",
    default=settings.MMMETA_FILES_ROOT,
    help="Base path for actual files to generate metadir from",
    show_default=True,
)
@click.option(
    "--log-level",
    default=settings.LOGGING,
    help="Set logging level",
    show_default=True,
)
@click.pass_context
def cli(ctx, metadir, files_root, log_level, invoke_without_command=True):
    configure_logging(log_level)
    if not metadir:
        raise click.BadParameter("Missing metadir root")
    if ctx.obj is None:
        ctx.obj = {}
    ctx.obj["m"] = mmmeta(metadir)


@cli.command()
@click.option(
    "--replace",
    is_flag=True,
    default=False,
    help="Completly replace the meta database",
    show_default=True,
)
@click.option(
    "--ensure",
    is_flag=True,
    default=False,
    help="Ensure metadata files are present, soft-delete non-existing",
    show_default=True,
)
@click.option(
    "--ensure-files",
    is_flag=True,
    default=False,
    help="Ensure actual files are present (for local store only), soft-delete non-existing",  # noqa
    show_default=True,
)
@click.option(
    "--no-meta",
    is_flag=True,
    default=False,
    help="Read in actual files instead of json metadata files",
    show_default=True,
)
@click.pass_context
def generate(ctx, replace, ensure, ensure_files, no_meta):
    path = None  # FIXME
    ctx.obj["m"].generate(path, replace, ensure, ensure_files, no_meta)


@cli.command()
@click.option(
    "--replace",
    is_flag=True,
    default=False,
    help="Completly replace the local state database",
    show_default=True,
)
@click.option(
    "--cleanup",
    is_flag=True,
    default=False,
    help="Try to do some data migrations, can be helpful when things break.",
    show_default=True,
)
@click.pass_context
def update(ctx, replace, cleanup):
    ctx.obj["m"].update(replace, cleanup)


@cli.command()
@click.pass_context
def inspect(ctx):
    meta = ctx.obj["m"]
    for key, value in meta.inspect().items():
        click.echo(f"{key}: {value}")
    click.echo(f"{meta.store.to_string()}")


@cli.command()
@click.pass_context
def squash(ctx):
    ctx.obj["m"].squash()


@cli.command()
@click.pass_context
def dump(ctx):
    ctx.obj["m"].dump()



import logging
from datetime import datetime
from itertools import chain
from pathlib import Path

import dataset
from banal import ensure_dict
from sqlalchemy.exc import IntegrityError

from .exceptions import ValidationError
from .util import casted_dict, checksum, dict_diff, dict_is_subset, robust_dict

log = logging.getLogger(__name__)


def _upsert(tx, metadir, files, prefix, ts, ensure=False, casted=False):
    # use explicit operations instead of upsert_many to be able to set some
    # more metadata
    table = tx["files"]
    to_insert = []
    updated = added = invalid = deleted = skipped = 0
    ignore_seen = set((("__seen", ts),))
    for file in files:
        if casted:
            file = casted_dict(file)
        if ensure:
            file["__seen"] = ts  # helper to do a quick scan later
        try:
            metadir.files.validate(file)
            uid = file[metadir.config.unique]
            existing_file = table.find_one(**{metadir.config.unique: uid})
            if existing_file:
                if dict_is_subset(file, existing_file, ignore_seen):
                    skipped += 1
                else:
                    if file.get("__deleted", None) is not None:
                        deleted += 1
                    else:
                        file[f"__{prefix}_last_updated"] = ts
                        updated += 1
                table.upsert(file, [metadir.config.unique])
            else:
                file[f"__{prefix}_added"] = ts
                file[f"__{prefix}_last_updated"] = ts
                to_insert.append(file)
                added += 1
        except ValidationError as e:
            invalid += 1
            fname = file.get(metadir.config.unique) or "undefined"
            log.error(f"File `{fname}` not valid: {e}")

    # at least bulk insert
    table.insert_many(to_insert)

    if ensure:
        for file in chain(
            table.find(__seen={"lt": ts}),
            table.find(__seen=None),
        ):
            file["__deleted"] = 1
            file["__deleted_at"] = ts
            file["__deleted_reason"] = f"{prefix}-missing"
            file[f"__{prefix}_last_updated"] = ts
            table.update(file, [metadir.config.unique])
            deleted += 1

    new_count = len(table)

    log.info(f"Updated {updated} files.")
    log.info(f"Added {added} new files.")
    log.info(f"Skipped {skipped} not changed files.")
    if invalid:
        log.warning(f"{invalid} invalid files")
    if deleted:
        log.warning(f"{deleted} soft deleted files")
    log.info(f"Now {new_count} files in database.")

    return updated, added, invalid, deleted, skipped


def _load_metadata(fp, metadir, ts, ensure_files=False):
    # use robust dict for performance
    keys = metadir.config.keys
    data = metadir._backend.load_json(fp)
    data = robust_dict(data)
    data = {k: v for k, v in data.items() if not keys or k in keys}
    if ensure_files:
        if not metadir.files.ensure(data):
            data["__meta_last_updated"] = ts
            data["__deleted"] = 1
            data["__deleted_at"] = ts
            data["__deleted_reason"] = "original-missing"
    return data


def _load_files(files):
    for _, fp in files:
        p = Path(fp)
        data = p.stat()
        yield {
            "file_name": p.name,
            "file_path": fp,
            "file_size": data.st_size,
            "created_at": datetime.fromtimestamp(data.st_ctime),
            "modified_at": datetime.fromtimestamp(data.st_mtime),
            "content_hash": checksum(fp),
        }


def _get_table(tx, primary_id, name="files"):
    return tx.get_table(name, primary_id=primary_id, primary_type=tx.types.text)


def update_state_db(metadir, replace=False, cleanup=False):
    """
    update remote metadata to local state
    """

    log.info(f"Updating metadata and state for `{metadir}` ...")

    with metadir._db as tx:
        if replace:
            # FIXME implement a soft delete? aka backup db file first
            tx["files"].drop()

        table = _get_table(tx, metadir.config.unique)
        primary_keys = table.table.primary_key.columns.keys()
        if metadir.config.unique not in primary_keys:
            # table was created before with another primary key
            # this is a bit hacky, but because of SQLite limitations,
            # we just make a new copy of the table with the new primary key...
            tmp_table = _get_table(tx, metadir.config.unique, "tmp")
            try:
                tmp_table.insert_many([i for i in table])
            except IntegrityError as e:
                log.warning(
                    f"Cannot perform `state.db` migration under such circumstances. Is your config correct? `{e}`"  # noqa
                )
            table.drop()
            table = _get_table(tx, metadir.config.unique)
            table.insert_many([i for i in tmp_table])
            tmp_table.drop()

        if cleanup:
            log.info("Cleaning up ...")
            tmp_table = _get_table(tx, metadir.config.unique, "tmp")
            keys = metadir.config.keys
            tmp_table.insert_many(
                [
                    casted_dict(
                        {
                            k: v
                            for k, v in f.items()
                            if not keys or k in keys or k.startswith("_")
                        }
                    )
                    for f in table
                ]
            )
            table = table.drop()
            table = _get_table(tx, metadir.config.unique)
            table.insert_many([i for i in tmp_table])

        log.info(f"{len(table)} exsiting files in `{tx}`")

        files = tx["meta_files"]
        metadir._metadata.load(files)
        # use a consistent timestamp for state diff queries
        ts = datetime.now()
        res = _upsert(tx, metadir, files, "state", ts, ensure=True, casted=True)
        files.drop()

    if any(res[:4]):
        # added or updated:
        metadir.touch("state_last_updated")
    return res


def generate_metadata(
    filebackend,
    metadir,
    replace=False,
    ensure_metadata=False,
    ensure_files=False,
    no_meta=False,
):
    """
    generate or update file metadata

    it is stored under `_mmmeta/db/*.json.meta` one file for each file

    ensure: soft delete all previously existing files not found in metadata
    """
    metadata = metadir._metadata

    if replace:
        log.warning(f"Replacing metadata for `{filebackend}` ...")
        metadata.delete()
    else:
        log.info(f"Updating metadata for `{filebackend}` ...")

    # use a consistent timestamp for state diff queries
    ts = datetime.now()

    # either read in json metadata files or actual files (only local filesystem here)
    if no_meta:
        files = _load_files(
            filebackend.get_children(condition=lambda x: "_mmmeta" not in x)
        )
    else:
        files = (
            _load_metadata(fp, metadir, ts, ensure_files)
            for _, fp in filebackend.get_children(
                condition=lambda x: x.endswith(".json")
            )
        )

    with dataset.connect("sqlite:///:memory:") as tx:
        metadb = _get_table(tx, metadir.config.unique)
        metadata.load(metadb)
        log.info(f"{metadb.count()} existing files.")

        # keep old state
        old_state = _get_table(tx, metadir.config.unique, "old_state")
        old_state.insert_many([i for i in metadb])

        res = _upsert(tx, metadir, files, "meta", ts, ensure_metadata)

        # export diff to append only db
        diff = _get_table(tx, metadir.config.unique, "diff")
        for file in metadb.find(__meta_last_updated=ts):
            old_file = ensure_dict(
                old_state.find_one(
                    **{metadir.config.unique: file[metadir.config.unique]}
                )
            )
            changed = dict(dict_diff(file, old_file))
            diff_data = {
                **changed,
                **{metadir.config.unique: file[metadir.config.unique]},
                **{"__mmmeta_keys": ",".join(changed.keys())},
            }
            diff.insert(diff_data)
        metadata.write(diff)

    if any(res[:4]):
        # added or updated:
        metadir.touch("meta_last_updated")

    return res
